"""Mixin extending Kili API Gateway class with Api Keys related operations."""

from typing import List, Optional, cast

from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.event.mappers import (
    event_pagination_wrapper,
    event_where_wrapper,
)
from kili.adapters.kili_api_gateway.event.operations import get_events_query
from kili.adapters.kili_api_gateway.event.queries import (
    PaginatedGraphQLQuery,
)
from kili.adapters.kili_api_gateway.helpers.queries import fragment_builder
from kili.domain.event import EventDict, EventFilters, QueryOptions
from kili.domain.types import ListOrTuple

DEFAULT_EVENT_FIELDS = [
    "createdAt",
    "event",
    "id",
    "organizationId",
    "payload",
    "projectId",
    "userId",
]


class EventOperationMixin(BaseOperationMixin):
    """Mixin extending Kili API Gateway class with event related operations."""

    def list_events(
        self,
        filters: EventFilters,
        options: QueryOptions,
        fields: Optional[ListOrTuple[str]] = None,
    ) -> List[EventDict]:
        """List events with given options."""
        fragment = fragment_builder(fields or DEFAULT_EVENT_FIELDS)
        query = get_events_query(fragment)
        where = event_where_wrapper(filters)
        pagination = event_pagination_wrapper(options)

        return [
            cast(EventDict, item)
            for item in PaginatedGraphQLQuery(
                self.graphql_client
            ).execute_query_from_paginated_call(
                query,
                where,
                pagination,
                options,
            )
        ]



"""Client presentation methods for labels."""

# pylint: disable=too-many-lines
from typing import List, Optional

from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway
from kili.domain.event import EventDict, EventFilters, OrderType, QueryOptions
from kili.domain.project import ProjectId
from kili.domain.user import UserId
from kili.utils.logcontext import for_all_methods, log_call


@for_all_methods(log_call, exclude=["__init__"])
class EventClientMethods:
    def __init__(self, kili_api_gateway: KiliAPIGateway):
        self.kili_api_gateway = kili_api_gateway

    def list(
        self,
        project_id: str,
        fields: Optional[List[str]] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        user_id: Optional[str] = None,
        event: Optional[str] = None,
        skip: int = 0,
        from_event_id: Optional[str] = None,
        until_event_id: Optional[str] = None,
        order: Optional[OrderType] = "asc",
    ) -> List[EventDict]:
        """List events of my project.

        Args:
            project_id: Identifier of the project.
            fields: All the fields to request among the possible fields for the events.
                Defaults is ["createdAt", "event", "id", "organizationId", "payload", "projectId", "userId"].
            created_at_gte: Filter events created after this date.
            created_at_lte: Filter events created before this date.
            user_id: Filter events by user id.
            event: Filter events by event type.
            skip: Number of events to skip.
            from_event_id: Filter events from this event id.
            until_event_id: Filter events until this event id.
            order: Order of the events. Can be "asc" or "desc".

        Returns:
            A list of events.

        Examples:
            >>> kili.events.list(project_id="your_project_id")
        """
        converted_filters = EventFilters(
            project_id=ProjectId(project_id),
            created_at_gte=created_at_gte,
            created_at_lte=created_at_lte,
            user_id=UserId(user_id) if user_id else None,
            event=event,
        )
        options = QueryOptions(
            skip=skip,
            from_event_id=from_event_id,
            until_event_id=until_event_id,
            order=order,
        )

        return list(
            self.kili_api_gateway.list_events(
                filters=converted_filters, options=options, fields=fields
            )
        )



"""GraphQL module."""

from typing import Any, Dict, Generator

from kili.core.graphql.graphql_client import GraphQLClient
from kili.domain.event import QueryOptions


class PaginatedGraphQLQuery:
    """Query class for querying Kili objects.

    It factorizes code for executing paginated queries.
    """

    def __init__(self, graphql_client: GraphQLClient) -> None:
        """Initialize the paginator."""
        self._graphql_client = graphql_client

    def execute_query_from_paginated_call(
        self,
        query: str,
        where: Dict[str, Any],
        pagination: Dict[str, Any],
        options: QueryOptions,
    ) -> Generator[Dict, None, None]:
        """Build a row generator from paginated query calls with the first and skip pattern.

        Args:
            query: The object query to execute and to send to graphQL, in string format
            where: The where payload to send in the graphQL query
            pagination: The where pagination payload to send in the graphQL query
            options: The query options with skip and first and disable_tqdm
        """
        count_elements_retrieved = 0
        while True:
            skip = count_elements_retrieved + options.skip
            first = options.batch_size
            order = options.order

            payload = {
                "where": where,
                "pagination": {"skip": skip, "first": first, **pagination},
                "order": order,
            }
            elements = self._graphql_client.execute(query, payload)["data"]
            if not isinstance(elements, list):
                raise TypeError(
                    "PaginatedGraphQLQuery only support operations returning a list of objects"
                )

            if len(elements) == 0:
                break

            yield from elements

            count_elements_retrieved += len(elements)

            if len(elements) < first:
                break



"""GraphQL Project Workflow operations."""


def get_update_project_workflow_mutation(fragment: str) -> str:
    """Return the GraphQL editProjectWorkflowSettings mutation."""
    return f"""
        mutation editProjectWorkflowSettings($input: EditProjectWorkflowSettingsInput!) {{
            data: editProjectWorkflowSettings(input: $input) {{
                {fragment}
            }}
        }}
        """



"""Types for the ProjectWorkflow-related Kili API gateway functions."""

from dataclasses import dataclass
from typing import Optional


@dataclass
class ProjectWorkflowDataKiliAPIGatewayInput:
    """ProjectWorkflow input data for Kili API Gateway."""

    enforce_step_separation: Optional[bool]



"""Event domain."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, NamedTuple, Optional

from typing_extensions import TypedDict

from kili.core.constants import QUERY_BATCH_SIZE

if TYPE_CHECKING:
    from .project import ProjectId
    from .user import UserId

OrderType = Literal["asc", "desc"]


@dataclass
class EventFilters:
    """Event filters for running an event search."""

    project_id: "ProjectId"
    created_at_gte: Optional[str] = None
    created_at_lte: Optional[str] = None
    user_id: Optional["UserId"] = None
    event: Optional[str] = None


class EventDict(TypedDict):
    """Dict that represents an Event."""

    created_at: str
    event: str
    id: str
    organization_id: str
    payload: dict
    project_id: str
    user_id: str


class QueryOptions(NamedTuple):
    """Options when calling GraphQLQuery from the SDK."""

    first: Optional[int] = None
    skip: int = 0
    batch_size: int = QUERY_BATCH_SIZE
    from_event_id: Optional[str] = None
    until_event_id: Optional[str] = None
    order: Optional[OrderType] = "asc"



"""GraphQL payload data mappers for project operations."""

from typing import Dict

from .types import ProjectWorkflowDataKiliAPIGatewayInput


def project_input_mapper(data: ProjectWorkflowDataKiliAPIGatewayInput) -> Dict:
    """Build the GraphQL ProjectWorfklowData variable to be sent in an operation."""
    return {
        "enforceStepSeparation": data.enforce_step_separation,
    }



"""Client presentation methods for project workflow."""

from typing import Any, Dict, Optional

from typeguard import typechecked

from kili.domain.project import ProjectId
from kili.use_cases.project_workflow import ProjectWorkflowUseCases

from .base import BaseClientMethods


class ProjectWorkflowClientMethods(BaseClientMethods):
    """Client presentation methods for project workflow."""

    @typechecked
    def update_project_workflow(
        self,
        project_id: str,
        enforce_step_separation: Optional[bool] = None,
    ) -> Dict[str, Any]:
        """Update properties of a project workflow.

        Args:
            project_id: Id of the project.
            enforce_step_separation: Prevents the same user from being assigned to
                multiple steps in the workflow for a same asset,
                ensuring independent review and labeling processes

        Returns:
            A dict with the changed properties which indicates if the mutation was successful,
                else an error message.
        """
        return ProjectWorkflowUseCases(self.kili_api_gateway).update_project_workflow(
            project_id=ProjectId(project_id),
            enforce_step_separation=enforce_step_separation,
        )



"""Mixin extending Kili API Gateway class with Projects related operations."""

from typing import Dict

from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.helpers.queries import (
    fragment_builder,
)
from kili.domain.project import ProjectId

from .mappers import project_input_mapper
from .operations import (
    get_update_project_workflow_mutation,
)
from .types import ProjectWorkflowDataKiliAPIGatewayInput


class ProjectWorkflowOperationMixin(BaseOperationMixin):
    """Mixin extending Kili API Gateway class with Projects workflow related operations."""

    def update_project_workflow(
        self,
        project_id: ProjectId,
        project_workflow_data: ProjectWorkflowDataKiliAPIGatewayInput,
    ) -> Dict:
        """Update properties in a project workflow."""
        project_workflow_input = project_input_mapper(data=project_workflow_data)

        fields = tuple(name for name, val in project_workflow_input.items() if val is not None)
        fragment = fragment_builder(fields)
        mutation = get_update_project_workflow_mutation(fragment)

        project_workflow_input["projectId"] = project_id

        variables = {"input": project_workflow_input}
        result = self.graphql_client.execute(mutation, variables)
        return result["data"]



"""GraphQL payload data mappers for api keys operations."""

from typing import Dict

from kili.domain.event import EventFilters, QueryOptions


def event_where_wrapper(filter: EventFilters) -> Dict:
    """Build the GraphQL EventMapperWhere variable to be sent in an operation."""
    return {
        "projectId": filter.project_id,
        "createdAtGte": filter.created_at_gte,
        "createdAtLte": filter.created_at_lte,
        "userId": filter.user_id,
        "event": filter.event,
    }


def event_pagination_wrapper(options: QueryOptions) -> Dict:
    """Build the GraphQL EventMapperPagination variable to be sent in an operation."""
    return {
        "fromEventId": options.from_event_id,
        "untilEventId": options.until_event_id,
    }



"""Project use cases."""

from typing import Dict, Optional

from kili.adapters.kili_api_gateway.project_workflow.types import (
    ProjectWorkflowDataKiliAPIGatewayInput,
)
from kili.domain.project import ProjectId
from kili.use_cases.base import BaseUseCases


class ProjectWorkflowUseCases(BaseUseCases):
    """ProjectWorkflow use cases."""

    def update_project_workflow(
        self,
        project_id: ProjectId,
        enforce_step_separation: Optional[bool] = None,
    ) -> Dict[str, object]:
        """Update properties in a project workflow."""
        project_workflow_data = ProjectWorkflowDataKiliAPIGatewayInput(
            enforce_step_separation=enforce_step_separation,
        )

        return self._kili_api_gateway.update_project_workflow(project_id, project_workflow_data)



"""Types used by the conversion service."""

from typing import Any, Dict, List, Literal, NamedTuple, NewType, Union

from typing_extensions import TypedDict

InputType = Literal["TEXT", "IMAGE"]

LogLevel = Union[int, Literal["ERROR", "WARNING", "DEBUG", "INFO", "CRITICAL"]]


class JobCategory(NamedTuple):
    """Contains information for a category."""

    category_name: str
    id: int
    job_id: str


MLTask = Literal["CLASSIFICATION", "NAMED_ENTITIES_RECOGNITION", "OBJECT_DETECTION"]
Tool = Literal["rectangle", "semantic", "polygon"]
JobName = NewType("JobName", str)


class Job(TypedDict):
    """Contains job settings."""

    content: Any
    instruction: str
    isChild: bool
    tools: List[Tool]
    mlTask: MLTask
    models: Any  # example: {"interactive-segmentation": {"job": "SEMANTIC_JOB_MARKER"}},
    isVisible: bool
    required: int
    isNew: bool


Jobs = Dict[JobName, Job]
